@@ -17,13 +17,13 @@ use tokio::time::timeout;
 use tokio::time::{Duration, Instant};
 use trust_dns_client::op::{Header, MessageType, OpCode};
 use trust_dns_client::proto::rr::rdata::TXT;
-use trust_dns_client::proto::rr::{DNSClass, RData, Record, RecordType};
+use trust_dns_client::proto::rr::{Name, RData, Record, RecordType};
 use trust_dns_server::authority::{MessageRequest, MessageResponse, MessageResponseBuilder};
 use trust_dns_server::server::{Request, RequestHandler, ResponseHandler};
 use trust_dns_server::ServerFuture;
 
 use common::dns::{
-    ClientId, ClientIdSuffixDecoder, Decoder, Encoder, HexDecoder, HexEncoder,
+    AppendSuffixDecoder, ClientId, ClientIdSuffixDecoder, Decoder, Encoder, HexDecoder, HexEncoder,
     CLIENT_ID_SIZE_IN_BYTES,
 };
 use common::io::Stream;
@@ -35,6 +35,7 @@ pub(crate) struct DnsUntunneler {
     listener_address: SocketAddr,
     read_timeout: Duration,
     idle_client_timeout: Duration,
+    client_suffix: String,
 }
 
 impl DnsUntunneler {
@@ -43,12 +44,14 @@ impl DnsUntunneler {
         local_port: u16,
         read_timeout: Duration,
         idle_client_timeout: Duration,
+        client_suffix: String,
     ) -> Result<Self, Box<dyn Error>> {
         let listener_address = SocketAddr::new(local_address, local_port);
         Ok(Self {
             listener_address,
             read_timeout,
             idle_client_timeout,
+            client_suffix,
         })
     }
 }
@@ -71,7 +74,11 @@ impl Untunneler for DnsUntunneler {
         );
 
         let udp_socket = UdpSocket::bind(self.listener_address).await?;
-        let mut server = ServerFuture::new(UntunnelRequestHandler::new(cache, self.read_timeout));
+        let mut server = ServerFuture::new(UntunnelRequestHandler::new(
+            cache,
+            self.read_timeout,
+            self.client_suffix.clone(),
+        ));
         server.register_socket(udp_socket);
         match server.block_until_done().await {
             Ok(_) => Ok(()),
@@ -101,13 +108,19 @@ const MAXIMUM_TXT_RECORD_SIZE: usize = 54;
 pub struct UntunnelRequestHandler<F: StreamCreator> {
     untunneled_clients: Arc<StreamsCache<F, ClientId>>,
     read_timeout: Duration,
+    client_suffix: String,
 }
 
 impl<F: StreamCreator> UntunnelRequestHandler<F> {
-    pub(crate) fn new(cache: StreamsCache<F, ClientId>, read_timeout: Duration) -> Self {
+    pub(crate) fn new(
+        cache: StreamsCache<F, ClientId>,
+        read_timeout: Duration,
+        client_suffix: String,
+    ) -> Self {
         Self {
             untunneled_clients: Arc::new(cache),
             read_timeout,
+            client_suffix,
         }
     }
 }
@@ -126,7 +139,10 @@ impl<F: StreamCreator> RequestHandler for UntunnelRequestHandler<F> {
                 handler: response_handle,
             },
             self.untunneled_clients.clone(),
-            ClientIdSuffixDecoder::new(HexDecoder {}),
+            AppendSuffixDecoder::new(
+                ClientIdSuffixDecoder::new(HexDecoder {}),
+                self.client_suffix.clone(),
+            ),
             HexEncoder {},
             self.read_timeout,
         ))
@@ -147,8 +163,13 @@ async fn untunnel_request<R: DnsResponseHandler, F: StreamCreator, D: Decoder, E
         None => return,
     };
 
-    log::debug!("received from tunnel {:?}", encoded_data_from_tunnel);
-    let data_from_tunnel = match decoder.decode(&encoded_data_from_tunnel) {
+    log::debug!(
+        "received from tunnel {}",
+        encoded_data_from_tunnel.to_string()
+    );
+    let mut non_fqdn_encoded_data = encoded_data_from_tunnel.clone();
+    non_fqdn_encoded_data.set_fqdn(false);
+    let data_from_tunnel = match decoder.decode(&non_fqdn_encoded_data.to_string()) {
         Ok(x) => x,
         Err(e) => {
             log::error!("{}: failed to decode: {}", message.id(), e);
@@ -177,11 +198,11 @@ async fn untunnel_request<R: DnsResponseHandler, F: StreamCreator, D: Decoder, E
         }
     };
     log::debug!("sending to tunnel {:?}", encoded_data_to_tunnel);
-    let answer = Record::new()
-        .set_rdata(RData::TXT(TXT::new(vec![encoded_data_to_tunnel])))
-        .set_record_type(RecordType::TXT)
-        .set_dns_class(DNSClass::IN)
-        .clone();
+    let answer = Record::from_rdata(
+        encoded_data_from_tunnel,
+        0,
+        RData::TXT(TXT::new(vec![encoded_data_to_tunnel])),
+    );
 
     let response = create_response(message, &answer);
     response_handle.send_response(response).unwrap_or_else(|e| {
@@ -189,7 +210,7 @@ async fn untunnel_request<R: DnsResponseHandler, F: StreamCreator, D: Decoder, E
     });
 }
 
-fn get_data_from_tunnel(message: &MessageRequest) -> Option<String> {
+fn get_data_from_tunnel(message: &MessageRequest) -> Option<Name> {
     let first_query = match message.queries().len() {
         1 => (&message.queries()[0]).original(),
         x => {
@@ -199,11 +220,7 @@ fn get_data_from_tunnel(message: &MessageRequest) -> Option<String> {
     };
 
     let data_from_tunnel = match first_query.query_type() {
-        RecordType::TXT => first_query
-            .name()
-            .to_string()
-            .trim_end_matches('.')
-            .to_string(),
+        RecordType::TXT => first_query.name().clone(),
         x => {
             log::error!("{}: unexpected type of query {}", message.id(), x);
             return None;
